@@ -9,16 +9,21 @@ import (
 	"github.com/tdewolff/parse/v2/buffer"
 )
 
+type Options struct {
+	WhileToFor bool
+}
+
 // Parser is the state for the parser.
 type Parser struct {
 	l   *Lexer
+	o   Options
 	err error
 
 	data                   []byte
 	tt                     TokenType
 	prevLT                 bool
 	inFor                  bool
-	async, generator       bool
+	await, yield           bool
 	assumeArrowFunc        bool
 	allowDirectivePrologue bool
 
@@ -29,11 +34,13 @@ type Parser struct {
 }
 
 // Parse returns a JS AST tree of.
-func Parse(r *parse.Input) (*AST, error) {
+func Parse(r *parse.Input, o Options) (*AST, error) {
 	ast := &AST{}
 	p := &Parser{
-		l:  NewLexer(r),
-		tt: WhitespaceToken, // trick so that next() works
+		l:     NewLexer(r),
+		o:     o,
+		tt:    WhitespaceToken, // trick so that next() works
+		await: true,
 	}
 
 	// process shebang
@@ -150,13 +157,7 @@ func (p *Parser) enterScope(scope *Scope, isFunc bool) *Scope {
 	parent := p.scope
 	p.scope = scope
 	*scope = Scope{
-		Parent:       parent,
-		Func:         nil,
-		Declared:     VarArray{},
-		Undeclared:   VarArray{},
-		NumVarDecls:  0,
-		NumArguments: 0,
-		HasWith:      false,
+		Parent: parent,
 	}
 	if isFunc {
 		scope.Func = scope
@@ -216,8 +217,8 @@ func (p *Parser) parseStmt(allowDeclaration bool) (stmt IStmt) {
 			return
 		}
 		p.next()
-		varDecl := p.parseVarDecl(tt)
-		stmt = &varDecl
+		varDecl := p.parseVarDecl(tt, true)
+		stmt = varDecl
 		if !p.prevLT && p.tt != SemicolonToken && p.tt != CloseBraceToken && p.tt != ErrorToken {
 			if tt == ConstToken {
 				p.fail("const declaration")
@@ -230,8 +231,7 @@ func (p *Parser) parseStmt(allowDeclaration bool) (stmt IStmt) {
 		let := p.data
 		p.next()
 		if allowDeclaration && (IsIdentifier(p.tt) || p.tt == YieldToken || p.tt == AwaitToken || p.tt == OpenBracketToken || p.tt == OpenBraceToken) {
-			varDecl := p.parseVarDecl(tt)
-			stmt = &varDecl
+			stmt = p.parseVarDecl(tt, false)
 			if !p.prevLT && p.tt != SemicolonToken && p.tt != CloseBraceToken && p.tt != ErrorToken {
 				p.fail("let declaration")
 				return
@@ -312,10 +312,22 @@ func (p *Parser) parseStmt(allowDeclaration bool) (stmt IStmt) {
 		if !p.consume("while statement", CloseParenToken) {
 			return
 		}
-		stmt = &WhileStmt{cond, p.parseStmt(false)}
+		body := p.parseStmt(false)
+		if p.o.WhileToFor {
+			varDecl := &VarDecl{TokenType: VarToken, Scope: p.scope, InFor: true}
+			p.scope.Func.VarDecls = append(p.scope.Func.VarDecls, varDecl)
+
+			block, ok := body.(*BlockStmt)
+			if !ok {
+				block = &BlockStmt{List: []IStmt{body}}
+			}
+			stmt = &ForStmt{varDecl, cond, nil, block}
+		} else {
+			stmt = &WhileStmt{cond, body}
+		}
 	case ForToken:
 		p.next()
-		await := p.async && p.tt == AwaitToken
+		await := p.await && p.tt == AwaitToken
 		if await {
 			p.next()
 		}
@@ -331,7 +343,7 @@ func (p *Parser) parseStmt(allowDeclaration bool) (stmt IStmt) {
 		if p.tt == VarToken || p.tt == LetToken || p.tt == ConstToken {
 			tt := p.tt
 			p.next()
-			varDecl := p.parseVarDecl(tt)
+			varDecl := p.parseVarDecl(tt, true)
 			if p.tt != SemicolonToken && (1 < len(varDecl.List) || varDecl.List[0].Default != nil) {
 				p.fail("for statement")
 				return
@@ -342,7 +354,7 @@ func (p *Parser) parseStmt(allowDeclaration bool) (stmt IStmt) {
 					return
 				}
 			}
-			init = &varDecl
+			init = varDecl
 		} else if p.tt != SemicolonToken {
 			init = p.parseExpression(OpExpr)
 		}
@@ -367,12 +379,19 @@ func (p *Parser) parseStmt(allowDeclaration bool) (stmt IStmt) {
 			if !p.consume("for statement", CloseParenToken) {
 				return
 			}
-			p.scope.MarkForInit()
+			p.scope.MarkForStmt()
 			if p.tt == OpenBraceToken {
 				body.List = p.parseStmtList("")
 			} else if p.tt != SemicolonToken {
 				body.List = []IStmt{p.parseStmt(false)}
 			}
+			if init == nil {
+				varDecl := &VarDecl{TokenType: VarToken, Scope: p.scope, InFor: true}
+				p.scope.Func.VarDecls = append(p.scope.Func.VarDecls, varDecl)
+				init = varDecl
+			} else if varDecl, ok := init.(*VarDecl); ok {
+				varDecl.InFor = true
+			}
 			stmt = &ForStmt{init, cond, post, body}
 		} else if p.tt == InToken {
 			if await {
@@ -384,12 +403,15 @@ func (p *Parser) parseStmt(allowDeclaration bool) (stmt IStmt) {
 			if !p.consume("for statement", CloseParenToken) {
 				return
 			}
-			p.scope.MarkForInit()
+			p.scope.MarkForStmt()
 			if p.tt == OpenBraceToken {
 				body.List = p.parseStmtList("")
 			} else if p.tt != SemicolonToken {
 				body.List = []IStmt{p.parseStmt(false)}
 			}
+			if varDecl, ok := init.(*VarDecl); ok {
+				varDecl.InForInOf = true
+			}
 			stmt = &ForInStmt{init, value, body}
 		} else if p.tt == OfToken {
 			p.next()
@@ -397,12 +419,15 @@ func (p *Parser) parseStmt(allowDeclaration bool) (stmt IStmt) {
 			if !p.consume("for statement", CloseParenToken) {
 				return
 			}
-			p.scope.MarkForInit()
+			p.scope.MarkForStmt()
 			if p.tt == OpenBraceToken {
 				body.List = p.parseStmtList("")
 			} else if p.tt != SemicolonToken {
 				body.List = []IStmt{p.parseStmt(false)}
 			}
+			if varDecl, ok := init.(*VarDecl); ok {
+				varDecl.InForInOf = true
+			}
 			stmt = &ForOfStmt{await, init, value, body}
 		} else {
 			p.fail("for statement", InToken, OfToken, SemicolonToken)
@@ -596,7 +621,7 @@ func (p *Parser) parseImportStmt() (importStmt ImportStmt) {
 		importStmt.Module = p.data
 		p.next()
 	} else {
-		if IsIdentifier(p.tt) || p.tt == YieldToken || p.tt == AwaitToken {
+		if IsIdentifier(p.tt) || p.tt == YieldToken {
 			importStmt.Default = p.data
 			p.next()
 			if p.tt == CommaToken {
@@ -609,7 +634,7 @@ func (p *Parser) parseImportStmt() (importStmt ImportStmt) {
 			if !p.consume("import statement", AsToken) {
 				return
 			}
-			if !IsIdentifier(p.tt) && p.tt != YieldToken && p.tt != AwaitToken {
+			if !IsIdentifier(p.tt) && p.tt != YieldToken {
 				p.fail("import statement", IdentifierToken)
 				return
 			}
@@ -617,18 +642,22 @@ func (p *Parser) parseImportStmt() (importStmt ImportStmt) {
 			p.next()
 		} else if p.tt == OpenBraceToken {
 			p.next()
-			for IsIdentifierName(p.tt) {
+			for IsIdentifierName(p.tt) || p.tt == StringToken {
+				tt := p.tt
 				var name, binding []byte = nil, p.data
 				p.next()
 				if p.tt == AsToken {
 					p.next()
-					if !IsIdentifier(p.tt) && p.tt != YieldToken && p.tt != AwaitToken {
+					if !IsIdentifier(p.tt) && p.tt != YieldToken {
 						p.fail("import statement", IdentifierToken)
 						return
 					}
 					name = binding
 					binding = p.data
 					p.next()
+				} else if !IsIdentifier(tt) && tt != YieldToken || tt == StringToken {
+					p.fail("import statement", IdentifierToken, StringToken)
+					return
 				}
 				importStmt.List = append(importStmt.List, Alias{name, binding})
 				if p.tt == CommaToken {
@@ -673,8 +702,8 @@ func (p *Parser) parseExportStmt() (exportStmt ExportStmt) {
 			p.next()
 			if p.tt == AsToken {
 				p.next()
-				if !IsIdentifierName(p.tt) {
-					p.fail("export statement", IdentifierToken)
+				if !IsIdentifierName(p.tt) && p.tt != StringToken {
+					p.fail("export statement", IdentifierToken, StringToken)
 					return
 				}
 				exportStmt.List = []Alias{Alias{star, p.data}}
@@ -688,13 +717,13 @@ func (p *Parser) parseExportStmt() (exportStmt ExportStmt) {
 			}
 		} else {
 			p.next()
-			for IsIdentifierName(p.tt) {
+			for IsIdentifierName(p.tt) || p.tt == StringToken {
 				var name, binding []byte = nil, p.data
 				p.next()
 				if p.tt == AsToken {
 					p.next()
-					if !IsIdentifierName(p.tt) {
-						p.fail("export statement", IdentifierToken)
+					if !IsIdentifierName(p.tt) && p.tt != StringToken {
+						p.fail("export statement", IdentifierToken, StringToken)
 						return
 					}
 					name = binding
@@ -726,8 +755,7 @@ func (p *Parser) parseExportStmt() (exportStmt ExportStmt) {
 	} else if p.tt == VarToken || p.tt == ConstToken || p.tt == LetToken {
 		tt := p.tt
 		p.next()
-		varDecl := p.parseVarDecl(tt)
-		exportStmt.Decl = &varDecl
+		exportStmt.Decl = p.parseVarDecl(tt, false)
 	} else if p.tt == FunctionToken {
 		exportStmt.Decl = p.parseFuncDecl()
 	} else if p.tt == AsyncToken { // async function
@@ -768,13 +796,18 @@ func (p *Parser) parseExportStmt() (exportStmt ExportStmt) {
 	return
 }
 
-func (p *Parser) parseVarDecl(tt TokenType) (varDecl VarDecl) {
+func (p *Parser) parseVarDecl(tt TokenType, canBeHoisted bool) (varDecl *VarDecl) {
 	// assume we're past var, let or const
-	varDecl.TokenType = tt
+	varDecl = &VarDecl{
+		TokenType: tt,
+		Scope:     p.scope,
+	}
 	declType := LexicalDecl
 	if tt == VarToken {
 		declType = VariableDecl
-		p.scope.Func.NumVarDecls++
+		if canBeHoisted {
+			p.scope.Func.VarDecls = append(p.scope.Func.VarDecls, varDecl)
+		}
 	}
 	for {
 		// binding element, var declaration in for-in or for-of can never have a default
@@ -789,6 +822,8 @@ func (p *Parser) parseVarDecl(tt TokenType) (varDecl VarDecl) {
 		} else if _, ok := bindingElement.Binding.(*Var); !ok && (!p.inFor || 0 < len(varDecl.List)) {
 			p.fail("var statement", EqToken)
 			return
+		} else if tt == ConstToken && (!p.inFor || p.inFor && p.tt != OfToken && p.tt != InToken) {
+			p.fail("const statement", EqToken)
 		}
 
 		varDecl.List = append(varDecl.List, bindingElement)
@@ -876,8 +911,8 @@ func (p *Parser) parseAnyFunc(async, inExpr bool) (funcDecl *FuncDecl) {
 		return
 	}
 	parent := p.enterScope(&funcDecl.Body.Scope, true)
-	parentAsync, parentGenerator := p.async, p.generator
-	p.async, p.generator = funcDecl.Async, funcDecl.Generator
+	parentAwait, parentYield := p.await, p.yield
+	p.await, p.yield = funcDecl.Async, funcDecl.Generator
 
 	if inExpr && name != nil {
 		funcDecl.Name, _ = p.scope.Declare(ExprDecl, name) // cannot fail
@@ -886,7 +921,7 @@ func (p *Parser) parseAnyFunc(async, inExpr bool) (funcDecl *FuncDecl) {
 	p.allowDirectivePrologue = true
 	funcDecl.Body.List = p.parseStmtList("function declaration")
 
-	p.async, p.generator = parentAsync, parentGenerator
+	p.await, p.yield = parentAwait, parentYield
 	p.exitScope(parent)
 	return
 }
@@ -1018,14 +1053,14 @@ func (p *Parser) parseClassElement() (method *MethodDecl, definition FieldDefini
 	}
 
 	parent := p.enterScope(&method.Body.Scope, true)
-	parentAsync, parentGenerator := p.async, p.generator
-	p.async, p.generator = method.Async, method.Generator
+	parentAwait, parentYield := p.await, p.yield
+	p.await, p.yield = method.Async, method.Generator
 
 	method.Params = p.parseFuncParams("method definition")
 	p.allowDirectivePrologue = true
 	method.Body.List = p.parseStmtList("method definition")
 
-	p.async, p.generator = parentAsync, parentGenerator
+	p.await, p.yield = parentAwait, parentYield
 	p.exitScope(parent)
 	return
 }
@@ -1072,7 +1107,7 @@ func (p *Parser) parseBindingElement(decl DeclType) (bindingElement BindingEleme
 
 func (p *Parser) parseBinding(decl DeclType) (binding IBinding) {
 	// binding identifier or binding pattern
-	if IsIdentifier(p.tt) || !p.generator && p.tt == YieldToken || !p.async && p.tt == AwaitToken {
+	if p.isIdentifierReference(p.tt) {
 		var ok bool
 		binding, ok = p.scope.Declare(decl, p.data)
 		if !ok {
@@ -1292,13 +1327,13 @@ func (p *Parser) parseObjectLiteral() (object ObjectExpr) {
 			if p.tt == OpenParenToken {
 				// MethodDefinition
 				parent := p.enterScope(&method.Body.Scope, true)
-				parentAsync, parentGenerator := p.async, p.generator
-				p.async, p.generator = method.Async, method.Generator
+				parentAwait, parentYield := p.await, p.yield
+				p.await, p.yield = method.Async, method.Generator
 
 				method.Params = p.parseFuncParams("method definition")
 				method.Body.List = p.parseStmtList("method definition")
 
-				p.async, p.generator = parentAsync, parentGenerator
+				p.await, p.yield = parentAwait, parentYield
 				p.exitScope(parent)
 				property.Value = &method
 				p.assumeArrowFunc = false
@@ -1394,10 +1429,10 @@ func (p *Parser) parseAsyncArrowFunc() (arrowFunc *ArrowFunc) {
 	// expect we're at Identifier or Yield or (
 	arrowFunc = &ArrowFunc{}
 	parent := p.enterScope(&arrowFunc.Body.Scope, true)
-	parentAsync, parentGenerator := p.async, p.generator
-	p.async, p.generator = true, false
+	parentAwait, parentYield := p.await, p.yield
+	p.await, p.yield = true, false
 
-	if IsIdentifier(p.tt) || !p.generator && p.tt == YieldToken {
+	if IsIdentifier(p.tt) || !p.yield && p.tt == YieldToken {
 		ref, _ := p.scope.Declare(ArgumentDecl, p.data)
 		p.next()
 		arrowFunc.Params.List = []BindingElement{{Binding: ref}}
@@ -1406,13 +1441,15 @@ func (p *Parser) parseAsyncArrowFunc() (arrowFunc *ArrowFunc) {
 
 		// could be CallExpression of: async(params)
 		if p.tt != ArrowToken {
+			// TODO?
 		}
 	}
 
+	p.await, p.yield = true, parentYield
 	arrowFunc.Async = true
 	arrowFunc.Body.List = p.parseArrowFuncBody()
 
-	p.async, p.generator = parentAsync, parentGenerator
+	p.await, p.yield = parentAwait, parentYield
 	p.exitScope(parent)
 	return
 }
@@ -1421,12 +1458,11 @@ func (p *Parser) parseIdentifierArrowFunc(v *Var) (arrowFunc *ArrowFunc) {
 	// expect we're at =>
 	arrowFunc = &ArrowFunc{}
 	parent := p.enterScope(&arrowFunc.Body.Scope, true)
-	parentAsync, parentGenerator := p.async, p.generator
-	p.async, p.generator = false, false
+	parentAwait, parentYield := p.await, p.yield
 
 	if 1 < v.Uses {
 		v.Uses--
-		v, _ = p.scope.Declare(ArgumentDecl, v.Data) // cannot fail
+		v, _ = p.scope.Declare(ArgumentDecl, parse.Copy(v.Data)) // cannot fail
 	} else {
 		// if v.Uses==1 it must be undeclared and be the last added
 		p.scope.Parent.Undeclared = p.scope.Parent.Undeclared[:len(p.scope.Parent.Undeclared)-1]
@@ -1434,10 +1470,11 @@ func (p *Parser) parseIdentifierArrowFunc(v *Var) (arrowFunc *ArrowFunc) {
 		p.scope.Declared = append(p.scope.Declared, v)
 	}
 
+	p.await = false
 	arrowFunc.Params.List = []BindingElement{{v, nil}}
 	arrowFunc.Body.List = p.parseArrowFuncBody()
 
-	p.async, p.generator = parentAsync, parentGenerator
+	p.await, p.yield = parentAwait, parentYield
 	p.exitScope(parent)
 	return
 }
@@ -1459,6 +1496,7 @@ func (p *Parser) parseArrowFuncBody() (list []IStmt) {
 	if p.tt == OpenBraceToken {
 		parentInFor := p.inFor
 		p.inFor = false
+		p.yield = false
 		p.allowDirectivePrologue = true
 		list = p.parseStmtList("arrow function")
 		p.inFor = parentInFor
@@ -1481,7 +1519,7 @@ func (p *Parser) parseAsyncExpression(prec OpPrec, async []byte) IExpr {
 	if !p.prevLT && p.tt == FunctionToken {
 		// primary expression
 		left = p.parseAsyncFuncExpr()
-	} else if !p.prevLT && prec <= OpAssign && (p.tt == OpenParenToken || IsIdentifier(p.tt) || !p.generator && p.tt == YieldToken || p.tt == AwaitToken) {
+	} else if !p.prevLT && prec <= OpAssign && (p.tt == OpenParenToken || IsIdentifier(p.tt) || !p.yield && p.tt == YieldToken || p.tt == AwaitToken) {
 		// async arrow function expression
 		if p.tt == AwaitToken {
 			p.fail("arrow function")
@@ -1606,11 +1644,11 @@ func (p *Parser) parseExpression(prec OpPrec) IExpr {
 		precLeft = OpUnary
 	case AwaitToken:
 		// either accepted as IdentifierReference or as AwaitExpression
-		if p.async && prec <= OpUnary {
+		if p.await && prec <= OpUnary {
 			p.next()
 			left = &UnaryExpr{tt, p.parseExpression(OpUnary)}
 			precLeft = OpUnary
-		} else if p.async {
+		} else if p.await {
 			p.fail("expression")
 			return nil
 		} else {
@@ -1677,7 +1715,7 @@ func (p *Parser) parseExpression(prec OpPrec) IExpr {
 		}
 	case YieldToken:
 		// either accepted as IdentifierReference or as YieldExpression
-		if p.generator && prec <= OpAssign {
+		if p.yield && prec <= OpAssign {
 			// YieldExpression
 			p.next()
 			yieldExpr := YieldExpr{}
@@ -1692,7 +1730,7 @@ func (p *Parser) parseExpression(prec OpPrec) IExpr {
 			}
 			left = &yieldExpr
 			precLeft = OpAssign
-		} else if p.generator {
+		} else if p.yield {
 			p.fail("expression")
 			return nil
 		} else {
@@ -2012,7 +2050,12 @@ func (p *Parser) parseExpressionSuffix(left IExpr, prec, precLeft OpPrec) IExpr
 				return left
 			}
 			p.next()
-			left = &BinaryExpr{tt, left, p.parseExpression(OpAssign)}
+			if commaExpr, ok := left.(*CommaExpr); ok {
+				commaExpr.List = append(commaExpr.List, p.parseExpression(OpAssign))
+				i-- // adjust expression nesting limit
+			} else {
+				left = &CommaExpr{[]IExpr{left, p.parseExpression(OpAssign)}}
+			}
 			precLeft = OpExpr
 		case ArrowToken:
 			// handle identifier => ..., where identifier could also be yield or await
@@ -2118,8 +2161,8 @@ func (p *Parser) parseParenthesizedExpressionOrArrowFunc(prec OpPrec, async []by
 	p.assumeArrowFunc, p.inFor = parentAssumeArrowFunc, parentInFor
 
 	if isArrowFunc {
-		parentAsync, parentGenerator := p.async, p.generator
-		p.async, p.generator = isAsync, false
+		parentAwait, parentYield := p.await, p.yield
+		p.await = isAsync
 
 		// arrow function
 		arrowFunc.Params = Params{List: make([]BindingElement, len(list))}
@@ -2130,7 +2173,7 @@ func (p *Parser) parseParenthesizedExpressionOrArrowFunc(prec OpPrec, async []by
 		arrowFunc.Params.Rest = p.exprToBinding(rest)
 		arrowFunc.Body.List = p.parseArrowFuncBody()
 
-		p.async, p.generator = parentAsync, parentGenerator
+		p.await, p.yield = parentAwait, parentYield
 		p.exitScope(parent)
 
 		left = arrowFunc
@@ -2159,11 +2202,11 @@ func (p *Parser) parseParenthesizedExpressionOrArrowFunc(prec OpPrec, async []by
 			precLeft = OpCall
 		} else {
 			// parenthesized expression
-			left = list[0]
-			for _, item := range list[1:] {
-				left = &BinaryExpr{CommaToken, left, item}
+			if 1 < len(list) {
+				left = &GroupExpr{&CommaExpr{list}}
+			} else {
+				left = &GroupExpr{list[0]}
 			}
-			left = &GroupExpr{left}
 		}
 	}
 	return p.parseExpressionSuffix(left, prec, precLeft)
@@ -2220,5 +2263,5 @@ func (p *Parser) exprToBindingElement(expr IExpr) (bindingElement BindingElement
 }
 
 func (p *Parser) isIdentifierReference(tt TokenType) bool {
-	return IsIdentifier(tt) || tt == YieldToken && !p.generator || tt == AwaitToken && !p.async
+	return IsIdentifier(tt) || tt == YieldToken && !p.yield || tt == AwaitToken && !p.await
 }
